#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define MAXBUFFER	512
#define BPARRAFO	16
#define MAXSEGMENTO	65535
#define MARCAEJEC   0x5643

typedef struct {
	unsigned marca;		/* Marca de ejecutable */
	unsigned tamcodigo;	/* Bytes del segmento de codigo */
	unsigned tamdatos;	/* Bytes de datos inicializados */
	unsigned tambbs;		/* Bytes de bbs */
	unsigned tamtotal;	/* Bytes totales asignados al programa */
	char reservado[22];
} CabPrograma;

int main(int argc, char *argv[])
{
	long tamcodigo, tamdatos, tambbs, tamtotal;
	FILE *forigen, *fdestino;
	CabPrograma cab;
	char buffer[MAXBUFFER];
	int num;

	/* Comprobacion numero de argumentos */
	if (argc != 6) {
		fprintf(stderr, "Uso: conv <tamcodigo> <tamdatos> <tambbs> <arc_origen> <arc_destino>\n");
		return 1;
	}

	/* Comprobacion de archivos */
	strlwr(argv[4]); strlwr(argv[5]);
	if (strcmp(argv[4], argv[5]) == 0) {
		fprintf(stderr, "Error: archivo destino igual a archivo origen\n");
		return 1;
	}

	/* Comprobacion de los valores de la memoria solicitada */
	tamcodigo = atol(argv[1]);
	tamdatos = atol(argv[2]);
	tambbs = atol(argv[3]);
	tamtotal = tamcodigo + tamdatos + tambbs;
	if ((tamtotal % BPARRAFO) != 0)
		tamtotal = (tamtotal/BPARRAFO)*BPARRAFO + BPARRAFO;
	if ((tamcodigo <= 0) || (tamcodigo > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para codigo incorrecta\n");
		return 1;
	}
	if ((tamdatos < 0) || (tamdatos > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para datos incorrecta\n");
		return 1;
	}
	if ((tambbs < 0) || (tambbs > MAXSEGMENTO)) {
		fprintf(stderr, "Memoria para bbs incorrecta\n");
		return 1;
	}
	if (tamtotal > MAXSEGMENTO) {
		fprintf(stderr, "Memoria total excede un segmento\n");
		return 1;
	}

	/* Apertura de archivos */
	if ((forigen = fopen(argv[4], "rb")) == NULL) {
		fprintf(stderr, "Error en archivo origen\n");
		return 1;
	}
	if ((fdestino = fopen(argv[5], "wb")) == NULL) {
		fclose(forigen);
		fprintf(stderr, "Error en archivo destino\n");
		return 1;
	}

	cab.marca = MARCAEJEC;
	cab.tamcodigo = tamcodigo;
	cab.tamdatos = tamdatos;
	cab.tambbs = tambbs;
	cab.tamtotal = tamtotal;
	if (fwrite(&cab, sizeof(cab), 1, fdestino) != 1) {
		fprintf(stderr, "Error en archivo destino\n");
		fclose(forigen); fclose(fdestino);
		return 1;
	}
	while (feof(forigen) == 0) {
		num = fread(buffer, 1, MAXBUFFER, forigen);
		fwrite(buffer, 1, num, fdestino);
	}
	fclose(forigen);
	fclose(fdestino);
	return 0;
}
