#ifndef MH_H
#define MH_H

#include "lista.h"
#include "memoria.h"

#define MHVERSION			"1.5"
#define FALSO				0
#define VERDAD				1
#define NUMINT				256
#define INT_RELOJ			8
#define INT_SISTEMA		134
#define TAM_PILA_HILO	8192
#define FLAGSINICIO		0x0200
#define HZRELOJ			100
#define TICKSQUANTUM		5
#define NUMMAXHILOS		25
#define NUMMINHILOS		6

#define MARCAPILA			0x4378
#define MARCAEJEC			0x5643
#define DESPLAZ_AX		16

/* Tipos de los hilos */
#define TNUCLEO		1
#define TUSUARIO		2

/* Estado de un hilo */
#define INVALIDO			1
#define EJECUCION			2
#define LISTO				3
#define DORMIDO			4
#define BLOQUEADO			5

/* Identificadores de las llamadas al sistema */
#define SNUEVOHILO  			1
#define STERMINARHILO		2
#define SFINHILO				3
#define SDORMIRHILO			4
#define SESCRIBIRCARACTER	5
#define SESCRIBIRCADENA		6
#define SLEERCARACTER		7

typedef const void (near *THiloFunc)(int Arg);
typedef struct {
	unsigned bp, di, si, ds, es, dx;
	unsigned cx, bx, ax, ip, cs, flags;
} TRegsPila;

typedef struct {
	int estado; 	/* Estado del hilo */
	int tipo;		/* Tipo de hilo: nucleo o usuario */
	int modo;		/* Modo de ejecucion: nucleo o usuario */
	unsigned ss;	/* Segmento de pila */
	unsigned isp;	/* Puntero inicial de pila */
	unsigned sp;	/* Puntero de pila en modo */
	int far *pmarca;	/* Puntero a la marca del segmento de pila */
	TRegsPila far *pregs;	/* Puntero a la estructura inicial de registros */
	THiloFunc pfunc;	/* Puntero a la funcion de un hilo del nucleo */
	int arg;	/* Argumento de la funcion de un hilo del nucleo */
	unsigned long diftick;	/* Diferencia de ticks para hilo dormido */
	int antid;	/* Indice del anterior hilo en la lista */
	int sigid;	/* Indice al hilo posterior en la lista */
} THilo;

typedef struct {
	unsigned marca;		/* Marca de ejecutable */
	unsigned tamcodigo;	/* Bytes del segmento de codigo */
	unsigned tamdatos;	/* Bytes de datos inicializados */
	unsigned tambbs;		/* Bytes de bbs */
	unsigned tamtotal;	/* Bytes totales asignados al programa */
	char reservado[22];
} TCabPrograma;

extern THilo thTablaHilos[NUMMAXHILOS];
extern int nHiloActual;
extern TLista tlHilosPreparados;
extern const long lCiclosPC;
extern long lHZReloj;
extern int nMaxHilos;
extern void far *AntTabla[256];

void interrupt ConmutarContexto(void);
void interrupt ManejadorReloj(void);
void interrupt InicioMultiHilo(void);
void interrupt LlamadaSistema(void);
void Planificador(void);
int InicializacionTablaHilos(void);
int NuevoPrograma(const char *archivo);
int NuevoHiloNucleo(THiloFunc ptfuncion, int Arg);

int MHNuevoHilo(THiloFunc ptfuncion, int narg, unsigned uiprologo);
int MHTerminarHilo(int nidilo);
int MHFinHilo(void);
int MHDormirHilo(unsigned long ulnumticks);

#endif
