#include <dos.h>
#include "consola.h"
#include "hilos.h"
#include "lista.h"
#include "memoria.h"
#include "varios.h"

THilo thTablaHilos[NUMMAXHILOS];
int nHiloActual = 1;
TLista tlHilosPreparados;
const long lCiclosPC = 1193180;
long lHZReloj = HZRELOJ;
int nMaxHilos;
void far *AntTabla[256];

static int nHiloAnterior = 1;
static int bPreventivo = VERDAD;
static int nNumHilos = 0;
static int nTickQuantum=0;
static unsigned long ulTickReloj = 0;
static TLista tlHilosDormidos;
static unsigned SegmentoNucleo;
static unsigned SegmentoPrograma;
static int nLlamada, nArg1, nArg2, nArg3, nResultado;

void MHFinMultiHilo(void)
{
	int i;

	InhibirInterrupciones();
	MHEscribirCadena("\nFin del sistema");

	/* Esperar pulsacion de teclado */
	while ((LeerBytePuerto(0x64) & 0x01) == 0) ;

	/* Desaceleracion del reloj hasta 18.2 ticks/segundo */
	EscribirBytePuerto(0x43, 0x34);
	EscribirBytePuerto(0x40, 0x00);
	EscribirBytePuerto(0x40, 0x00);

	/* Instalacion de la anterior tabla de interrupciones */
#ifndef DEBUG
	for (i=0; i<256; i++) EscribirVectorInt(i, AntTabla[i]);
#else
	EscribirVectorInt(INT_RELOJ, AntTabla[INT_RELOJ]);
	EscribirVectorInt(INT_TECLADO, AntTabla[INT_TECLADO]);
	EscribirVectorInt(INT_SISTEMA, AntTabla[INT_SISTEMA]);
#endif

	/* Activar todas las interrupciones hardware */
	EscribirBytePuerto(0x21, 0x00);
	EscribirBytePuerto(0XA1, 0x00);

	/* Regresar al MS-DOS */
	asm {
		mov ah, 4Ch
		int 21h
	}
}

static int Cargador(const char *archivo)
{
	union REGS rin, rout;
	struct SREGS segs;
	int fin = 0;
	int f;
	TCabPrograma cab;
	unsigned num;
	char far *Memoria;
	int nflags;

	nflags = DesactivarInt();
	/* Apertura del archivo */
	rin.h.ah = 0x3D;	/* Funcion de apertura de archivo */
	rin.h.al = 0;		/* Modo de apertura: lectura */
	rin.x.dx = FP_OFF(archivo); /* Nombre del archivo: Desplazamiento */
	segs.ds = FP_SEG(archivo); /* Nombre archivo: segmento */
	int86x(0x21, &rin, &rout, &segs);
	if (rout.x.cflag != 0) {
		RecuperarFlags(nflags);
		return 1;
	}
	f = rout.x.ax; /* Manejador del archivo */

	/* Lectura cabecera del archivo */
	rin.h.ah = 0x3F;	/* Funcion de lectura de archivo */
	rin.x.bx = f;		/* Manejador del archivo */
	rin.x.cx = sizeof(TCabPrograma);
	rin.x.dx = FP_OFF(&cab); /* Buffer: Desplazamiento */
	segs.ds = FP_SEG(&cab); /* Buffer: segmento */
	int86x(0x21, &rin, &rout, &segs);
	if (rout.x.ax != sizeof(TCabPrograma)) {
		rin.h.ah = 0x3E;	/* Funcion de cierre de archivo */
		rin.x.bx = f; /* Manejador del archivo */
		int86(0x21, &rin, &rout);
		RecuperarFlags(nflags);
		return 3;
	}

	if (cab.marca != MARCAEJEC) {
		rin.h.ah = 0x3E;	/* Funcion de cierre de archivo */
		rin.x.bx = f; /* Manejador del archivo */
		int86(0x21, &rin, &rout);
		RecuperarFlags(nflags);
		return 4;
	}

	if ((SegmentoPrograma = AsignarMemoria(cab.tamtotal+0x0100)) == 0) {
		rin.h.ah = 0x3E;	/* Funcion de cierre de archivo */
		rin.x.bx = f; /* Manejador del archivo */
		int86(0x21, &rin, &rout);
		RecuperarFlags(nflags);
		return 5;
	}

	Memoria = MK_FP(SegmentoPrograma, 0x0100);

	/* Carga del codigo y datos inicializados del programa */
	num = cab.tamtotal;
	while ((fin == 0) && (num > 0)) {
		rin.h.ah = 0x3F;	/* Funcion de lectura de archivo */
		rin.x.bx = f;		/* Manejador del archivo */
		if (num >= 512) { /* Numero de bytes a leer */
			rin.x.cx = 512;
		}
		else rin.x.cx = num;
		rin.x.dx = FP_OFF(Memoria); /* Buffer: Desplazamiento */
		segs.ds = FP_SEG(Memoria); /* Buffer: segmento */
		int86x(0x21, &rin, &rout, &segs);
		Memoria += rin.x.cx; num -= rin.x.cx;
		if (rout.x.ax < rin.x.cx) fin = 1;
	}

	/* Cierre del archivo */
	rin.h.ah = 0x3E;	/* Funcion de cierre de archivo */
	rin.x.bx = f; /* Manejador del archivo */
	int86(0x21, &rin, &rout);
	if (fin == 1) {
		RecuperarFlags(nflags);
		return 0;
	}
	RecuperarFlags(nflags);
	return 2;
}

void interrupt ManejadorReloj(void)
{
	int i, id;

	ulTickReloj++; nTickQuantum++;
	/* Comprobacion de hilos dormidos para despertar */
	for (i = tlHilosDormidos.cabeza; i != -1; i = thTablaHilos[i].sigid) {
		if (thTablaHilos[i].diftick > 1) {
			thTablaHilos[i].diftick--;
			break;
		}
		id = ExtraerHilo(&tlHilosDormidos);
		if (id >= 0) { /* En caso contrario, error interno del sistema */
			thTablaHilos[id].estado = LISTO;
			IntroducirHilo(id, &tlHilosPreparados);
		}
	}

	EscribirBytePuerto(0x20, 0x20);

	/* Comprobacion de destruccion de pila del hilo actual */
	if (*thTablaHilos[nHiloActual].pmarca != MARCAPILA) {
		MHEscribirCadena("Desbordamiento de pila");
		Planificador();
		/* Conmutacion de contexto */
		thTablaHilos[nHiloAnterior].sp = _SP;
		_SS = thTablaHilos[nHiloActual].ss;
		_SP = thTablaHilos[nHiloActual].sp;
	}

	/* Conmutacion de hilo si el sistema es preventivo y procede */
	if ((bPreventivo == VERDAD) && (nTickQuantum >= TICKSQUANTUM)) {
		Planificador();
		/* Conmutacion de contexto */
		thTablaHilos[nHiloAnterior].sp = _SP;
		_SS = thTablaHilos[nHiloActual].ss;
		_SP = thTablaHilos[nHiloActual].sp;
	}
}

void interrupt InicioMultiHilo(void)
{
	_SS = thTablaHilos[1].ss;
	_SP = thTablaHilos[1].sp;
}

void interrupt ConmutarContexto(void)
{
	thTablaHilos[nHiloAnterior].sp = _SP;
	_SS = thTablaHilos[nHiloActual].ss;
	_SP = thTablaHilos[nHiloActual].sp;
}

void Planificador(void)
{
	int nflags;

	nflags = DesactivarInt();
	if (nNumHilos <= 1) {
		MHFinMultiHilo(); /* Llamada sin retorno */
	}

	if (thTablaHilos[nHiloActual].estado == EJECUCION) {
		thTablaHilos[nHiloActual].estado = LISTO;
		IntroducirHilo(nHiloActual, &tlHilosPreparados);
	}

	nHiloAnterior = nHiloActual;
	nHiloActual = ExtraerHilo(&tlHilosPreparados);
	thTablaHilos[nHiloActual].estado = EJECUCION;
	nTickQuantum = 0;
	RecuperarFlags(nflags);
}

static void LlamadasSistema1(void)
{
	void far *p;
	unsigned long num;

	switch (nLlamada) {
		case SNUEVOHILO:
			nResultado = MHNuevoHilo((THiloFunc)nArg1, nArg2, nArg3);
			break;

		case STERMINARHILO:
			nResultado = MHTerminarHilo(nArg1);
			break;

		case SFINHILO:
			nResultado = MHFinHilo();
			break;

		case SDORMIRHILO:
			num = (nArg2 << 8) + nArg1;
			nResultado = MHDormirHilo(num);
			break;

		case SESCRIBIRCARACTER:
			nResultado = MHEscribirCaracter((char)nArg1);
			break;

		case SESCRIBIRCADENA:
			p = MK_FP(SegmentoPrograma, nArg1);
			nResultado = MHEscribirCadena((char far *)p);
			break;

		case SLEERCARACTER:
			nResultado = MHLeerCaracter();
			break;

		default: MHEscribirCadena("Error en llamada de sistema");
					nResultado = -1;
					break;
	}
	return;
}

void interrupt LlamadaSistema(void)
{
	/* Argumentos de la llamada al sistema */
	nLlamada = _AX; nArg1 = _BX; nArg2 = _CX; nArg3 = _DX;

	/* Comprobacion de destruccion de pila del hilo actual */
	if (*thTablaHilos[nHiloActual].pmarca != MARCAPILA) {
		MHEscribirCadena("Desbordamiento de pila");
		Planificador();
		ConmutarContexto();
	}

	/* Invocar llamada al sistema */
	LlamadasSistema1();

	/* Valor de la llamada al sistema */
	asm mov ax, nResultado
	asm mov [bp+DESPLAZ_AX], ax
}

static int NuevoIdentHilo(void)
{
	int i, nflags;

	nflags = DesactivarInt();
	for (i=0; i<nMaxHilos; i++) {
		if (thTablaHilos[i].estado == INVALIDO) break;
	}
	if (i == nMaxHilos) {
		RecuperarFlags(nflags);
		return -1;
	}
	RecuperarFlags(nflags);
	return i;
}

int NuevoPrograma(const char *archivo)
{
	int id, nflags, err;

	nflags = DesactivarInt();
	if ((id = NuevoIdentHilo()) == -1) {
		RecuperarFlags(nflags);
		return -1;
	}
	if ((err = Cargador(archivo)) != 0) {
		RecuperarFlags(nflags);
		return -err;
	}

	thTablaHilos[id].estado = LISTO;
	thTablaHilos[id].tipo = TUSUARIO;
	thTablaHilos[id].sp = thTablaHilos[id].isp;
	thTablaHilos[id].pfunc = 0;
	thTablaHilos[id].arg = 0;
	thTablaHilos[id].pregs->cs = SegmentoPrograma;
	thTablaHilos[id].pregs->ds = SegmentoPrograma;
	thTablaHilos[id].pregs->es = SegmentoPrograma;
	thTablaHilos[id].pregs->ip = 0x0100;
	thTablaHilos[id].pregs->flags = FLAGSINICIO;
	thTablaHilos[id].diftick = 0;
	IntroducirHilo(id, &tlHilosPreparados);
	nNumHilos++;
	RecuperarFlags(nflags);
	return id;
}

static void ShellHilo(void)
{
	(*(thTablaHilos[nHiloActual].pfunc))(thTablaHilos[nHiloActual].arg);
	MHFinHilo();
}

int NuevoHiloNucleo(THiloFunc pFuncion, int Arg)
{
	int id, nflags;

	nflags = DesactivarInt();
	if ((id = NuevoIdentHilo()) == -1) {
		RecuperarFlags(nflags);
		return -1;
	}

	thTablaHilos[id].estado = LISTO;
	thTablaHilos[id].tipo = TNUCLEO;
	thTablaHilos[id].sp = thTablaHilos[id].isp;
	thTablaHilos[id].pfunc = pFuncion;
	thTablaHilos[id].arg = Arg;
	thTablaHilos[id].pregs->cs = SegmentoNucleo;
	thTablaHilos[id].pregs->ds = SegmentoNucleo;
	thTablaHilos[id].pregs->es = SegmentoNucleo;
	thTablaHilos[id].pregs->ip = FP_OFF(ShellHilo);
	thTablaHilos[id].pregs->flags = FLAGSINICIO;
	thTablaHilos[id].diftick = 0;
	IntroducirHilo(id, &tlHilosPreparados);
	nNumHilos++;
	RecuperarFlags(nflags);
	return id;
}

int MHNuevoHilo(THiloFunc pFuncion, int Arg, unsigned uiPrologo)
{
	int id, nflags;
	int far *pnPilaAnterior;

	nflags = DesactivarInt();
	if ((id = NuevoIdentHilo()) == -1) {
		RecuperarFlags(nflags);
		return -1;
	}

	thTablaHilos[id].estado = LISTO;
	thTablaHilos[id].tipo = TUSUARIO;
	thTablaHilos[id].sp = thTablaHilos[id].isp;
	thTablaHilos[id].pfunc = pFuncion;
	thTablaHilos[id].arg = Arg;
	thTablaHilos[id].pregs->cs = SegmentoPrograma;
	thTablaHilos[id].pregs->ds = SegmentoPrograma;
	thTablaHilos[id].pregs->es = SegmentoPrograma;
	thTablaHilos[id].pregs->ip = uiPrologo;
	thTablaHilos[id].pregs->flags = FLAGSINICIO;
	thTablaHilos[id].diftick = 0;
	pnPilaAnterior = (int far *)((char far *)thTablaHilos[id].pregs
		+ sizeof(TRegsPila));
	*pnPilaAnterior = 0; /* Direccion de retorno de funcion anterior*/
	pnPilaAnterior++;
	*pnPilaAnterior = Arg; /* Argumento */
	pnPilaAnterior++;
	*pnPilaAnterior = (int)pFuncion;
	IntroducirHilo(id, &tlHilosPreparados);
	nNumHilos++;
	RecuperarFlags(nflags);
	return id;
}

int MHDormirHilo(unsigned long NumTicks)
{
	int act, nflags;

	nflags = DesactivarInt();
	thTablaHilos[nHiloActual].estado = DORMIDO;
	for (act = tlHilosDormidos.cabeza; act != -1; act = thTablaHilos[act].sigid) {
		if (NumTicks < thTablaHilos[act].diftick) { /* Inserto delante de este */
			thTablaHilos[nHiloActual].diftick = NumTicks;
			thTablaHilos[nHiloActual].sigid = act;
			if (thTablaHilos[act].antid != -1) {
				thTablaHilos[nHiloActual].antid = thTablaHilos[act].antid;
				thTablaHilos[thTablaHilos[act].antid].sigid = nHiloActual;
				thTablaHilos[act].antid = nHiloActual;
				break;
			}
			else { /* Primero de la lista */
				thTablaHilos[nHiloActual].antid = -1;
				thTablaHilos[act].antid = nHiloActual;
				tlHilosDormidos.cabeza = nHiloActual;
			}
		}
		if (thTablaHilos[act].sigid == -1) { /* Inserto al final de la lista */
			thTablaHilos[act].sigid = nHiloActual;
			thTablaHilos[nHiloActual].antid = act;
			thTablaHilos[nHiloActual].sigid = -1;
			tlHilosDormidos.cola = nHiloActual;
			thTablaHilos[nHiloActual].diftick = NumTicks;
			break;
		}
		NumTicks -= thTablaHilos[act].diftick;
	}
	/* Inserto al inicio de la lista */
	if (act == -1) {
		tlHilosDormidos.cabeza = tlHilosDormidos.cola = nHiloActual;
		thTablaHilos[nHiloActual].antid = thTablaHilos[nHiloActual].sigid = -1;
		thTablaHilos[nHiloActual].diftick = NumTicks;
	}
	Planificador();
	ConmutarContexto();
	RecuperarFlags(nflags);
	return 0;
}

int MHTerminarHilo(int nidhilo)
{
	int nflags;

	nflags = DesactivarInt();
	if ((nidhilo >= nMaxHilos) || (nidhilo <= 0)) {
		RecuperarFlags(nflags);
		return -1;
	}
	switch (thTablaHilos[nidhilo].estado) {
		case LISTO:			BorrarHilo(nidhilo, &tlHilosPreparados);
								break;
		case DORMIDO:		BorrarHilo(nidhilo, &tlHilosDormidos);
								break;
		case BLOQUEADO:	BorrarHilo(nidhilo, &tlHilosBloqueados);
								break;
	}
	thTablaHilos[nidhilo].estado = INVALIDO;
	nNumHilos--;
	if (nidhilo == nHiloActual) {
		Planificador();
		ConmutarContexto();
	}
	RecuperarFlags(nflags);
	return 0;
}

int MHFinHilo(void)
{
	return MHTerminarHilo(nHiloActual);
}

int InicializacionTablaHilos(void)
{
	int i;

	SegmentoNucleo = _CS;
	for (i=0; i<NUMMAXHILOS; i++, nMaxHilos++) {
		thTablaHilos[i].estado = INVALIDO;
		thTablaHilos[i].tipo = TUSUARIO;
		thTablaHilos[i].ss = AsignarMemoria(TAM_PILA_HILO);
		if (thTablaHilos[i].ss == 0) break;
		thTablaHilos[i].isp = TAM_PILA_HILO - 10*sizeof(int) - sizeof(TRegsPila);
		thTablaHilos[i].sp = thTablaHilos[i].isp;
		thTablaHilos[i].pmarca = (int far *)MK_FP(thTablaHilos[i].ss,
			TAM_PILA_HILO - sizeof(int));
		*thTablaHilos[i].pmarca = MARCAPILA;
		thTablaHilos[i].pregs = (TRegsPila far *)MK_FP(thTablaHilos[i].ss,
			TAM_PILA_HILO - 10*sizeof(int) - sizeof(TRegsPila));
		thTablaHilos[i].pfunc = 0;
		thTablaHilos[i].arg = 0;
		thTablaHilos[i].sigid = thTablaHilos[i].antid = -1;
	}

	tlHilosPreparados.cabeza = tlHilosPreparados.cola = -1;
	tlHilosDormidos.cabeza = tlHilosDormidos.cola = -1;
	tlHilosBloqueados.cabeza = tlHilosBloqueados.cola = -1;

	if (nMaxHilos < NUMMINHILOS) return 1;
	return 0;
}
