#include "consola.h"
#include "hilos.h"
#include "varios.h"

void TareaNula(void)
{
	int nflags;

	while (VERDAD) { /* Bucle infinito */
		nflags = DesactivarInt();
		Planificador(); /* Selecciona otro hilo */
		ConmutarContexto();
		RecuperarFlags(nflags);
	}
}

void interrupt Excepcion(void)
{
	MHEscribirCadena("Interrupcion no esperada. Pulse cualquier tecla");
	MHLeerCaracter();
	MHFinHilo();
}

void main(void)
{
	int i, nerr;
	unsigned uidivisor;

#ifndef __TINY__
#error Este programa solo funciona en el modelo de memoria Tiny
#endif
	InhibirInterrupciones();

	InicializacionConsola();
	MHEscribirCadena("Iniciando MH "); MHEscribirCadena(MHVERSION);
	EscribirCadenaNumero("\nTicks de reloj por segundo: ", HZRELOJ);
	EscribirCadenaNumero("\nQuantum en ticks de reloj: ", TICKSQUANTUM);
	MHEscribirCaracter('\n');

	if (InicializacionMemoria() != 0) {
		MHEscribirCadena("Error en inicializacion de memoria");
		return;
	}
	EscribirCadenaNumero("Memoria disponible: ", NumParrafosInicial/64);
	MHEscribirCadena(" Kbytes\n");

	if (InicializacionTablaHilos() != 0) {
		MHEscribirCadena("Error en inicializacion de tabla de hilos");
		return;
	}
	EscribirCadenaNumero("Numero max. hilos: ", nMaxHilos);
	EscribirCadenaNumero(" (", TAM_PILA_HILO);
	MHEscribirCadena(" bytes de pila por hilo)");
	MHEscribirCaracter('\n');

	if ((NuevoHiloNucleo((THiloFunc)TareaNula, 0)) != 0) {
		MHEscribirCadena("Error en tarea nula\n");
		return;
	}

	if (((nerr = NuevoPrograma("init")) < 1)) {
		EscribirCadenaNumero("Error en la carga del init: ", nerr);
		return;
	}

	ApagarMotoresDisketes();

	/* Se guarda la tabla de interrupciones anterior */
	for (i=0; i<256; i++) AntTabla[i] = LeerVectorInt(i);

#ifndef DEBUG
	for (i=0; i<256; i++) EscribirVectorInt(i, (void far *)Excepcion);
#endif

	/* Inhibir interrupciones hardware no deseadas */
	EscribirBytePuerto(0x21, 0xFC); /* Solo reloj y teclado */
	EscribirBytePuerto(0XA1, 0xFF);

	/* Aceleracion del reloj a hz ticks por segundo */
	uidivisor = lCiclosPC / lHZReloj;
	EscribirBytePuerto(0x43, 0x34);
	EscribirBytePuerto(0x40, uidivisor & 0x00FF);
	EscribirBytePuerto(0x40, (uidivisor >> 8) & 0x00FF);

	/* Instalacion del manejador del reloj */
	EscribirVectorInt(INT_RELOJ, (void far *)ManejadorReloj);

	/* Instalacion del manejador del teclado */
	EscribirVectorInt(INT_TECLADO, (void far *)ManejadorTeclado);

	/* Instalacion del manejador del llamadas al sistema */
	EscribirVectorInt(INT_SISTEMA, (void far *)LlamadaSistema);

	MHEscribirCadena("Inicio multihilo\n");
	InicioMultiHilo(); /* Inicio del sistema multihilo */
}
